use osu_file_parser::{
    osu_file::{
        hitobjects::{
            CurveType::*,
            HitObjectParams::*
        }
    },
    OsuFile
};
use rust_decimal::{Decimal, prelude::ToPrimitive};

mod types;
use types::*;

/*
    The input file.
*/
const OSU: &str = include_str!("../osu/test2/test2.osu");

/*
    Machine's maximum acceleration in mm/s².
*/
const MAX_ACCEL: f64 = 110_000.0;
/*
    The default acceleration to use for max_v calculation.
*/
const DEFAULT_ACCEL: f64 = MAX_ACCEL / 2.0;
/*
    The acceleration to use for groups of hitobjects. This should be higher as
    groups of hitobjects usually have to be traveled with "linear" velocity
    across all the hitobjects.
*/
const GROUP_ACCEL: f64 = MAX_ACCEL;
/*
    How much to increase the acceleration if the current acceleration doesn't
    produce a result.
*/
const ACCEL_RETRY_STEP: f64 = 5_000.0;
/*
    Since hitobjects' times are rounded to the nearest millisecond, linear
    velocity calculations aren't very precise, leading to hitobjects with linear
    velocities perceived to be identical not counted for grouping. To compensate
    for this, a threshold is set, and any difference between linear velocities
    under this threshold are counted.
*/
const GROUP_VEL_LENIENCY: f64 = 0.2;
 
/*
    Size of the bounding box for the beatmap in storage. Do not change.
*/
const BOUNDING_BOX_SIZE: (f64, f64) = (512.0, 384.0); // lazer default ig
/*
    Physical limits of the machine
*/
const PHYSICAL_MAX: (f64, f64) = (105.0, 110.0);
/*
    Active area
*/
const AREA_ORIGIN: (f64, f64) = (10.0, 10.0);
const AREA_MAX: (f64, f64) = (95.0, 95.0);
/*
    Rotation in degrees anteclockwise
*/
const ROTATION: f64 = 90.0;

const AREA_SIZE: (f64, f64) = (
    AREA_MAX.0 - AREA_ORIGIN.0,
    AREA_MAX.1 - AREA_ORIGIN.1
);
const BOX_TO_AREA_OFFSET: (f64, f64) = (
    AREA_SIZE.0 / 2.0 - BOUNDING_BOX_SIZE.0 / 2.0,
    AREA_SIZE.1 / 2.0 - BOUNDING_BOX_SIZE.1 / 2.0
);

/*
    Calculates the maximum velocity (mm/ms) given the the start and end
    position, acceleration, and required total move time.

    This uses an equation from 19aksh https://math.stackexchange.com/a/3304761
    rearranged to solve for v.
    
    In loose mathematical format:

        2t - sqrt(4t² - 8d (1 / a - 1 / -a))
    v = -----------------------------------
                2 (1 / a - 1 / -a)

    Where:
        t = time (ms)
        d = distance (mm)
        a = kinematic acceleration (mm/1000² since 1s == 1000ms)
        v = velocity (mm/ms)

    (multiply v by 60_000 to get the feedrate (mm/min))

    There are some combinations of t, d, and a that can result in no solution,
    usually because the acceleration and deceleration periods overlap. In this
    case, the acceleration should be increased until a solution is found. To
    compensate for the possibility of this exceeding the MAX_ACCEL, this will
    normally calculate with half of MAX_ACCEL. If no solution is found even
    at MAX_ACCEL, an error will be thrown.
 */
fn max_velocity(
    p1: PositionF64,
    p2: PositionF64,
    accel: f64,
    ms: f64,
) -> f64 {
    let dist = p1 >> p2;
    let accel_divided = accel / 1_000_000.0; // accel / 1000² since 1s == 1000ms 
    let accel_factor = 1.0 / accel_divided - 1.0 / -accel_divided;
    let sqrt_fac = (4.0 * ms.powf(2.0) - 8.0 * dist * accel_factor).sqrt();
    (2.0 * ms - sqrt_fac) / (2.0 * accel_factor)
}

fn find_bbox_scale_factor(
    box_size: (f64, f64),
    target_size: (f64, f64),
    rotation: f64
) -> f64 {
    let target_size = if rotation == 90.0 || rotation == 270.0 {
        (target_size.1, target_size.0)
    } else {
        target_size
    };
    if target_size.0 > target_size.1 {
        target_size.1 / box_size.1
    } else {
        target_size.0 / box_size.0
    }
}

fn solve_feedrate_and_accel(
    p1: PositionF64,
    p2: PositionF64,
    starting_accel: f64,
    ms: f64
) -> (f64, f64) {
    let mut accel = starting_accel;
    let mut max_v: f64;

    loop {
        max_v = max_velocity(p1, p2, accel, ms);
        if !max_v.is_nan() { break }
        accel += ACCEL_RETRY_STEP;
        if accel >= MAX_ACCEL {
            panic!("couldn't calculate max_v for dist {} within {}ms", p1 >> p2, ms);
        }
    }

    let speed = max_v * 1_000.0; // mm/s
    let feedrate = speed * 60.0; // mm/min

    eprintln!(
        ";   feedrate: {:.2} ({:.2}mm/s) ({})->({}) dist {}, accel={}, Δms={}",
        feedrate, speed,
        p1, p2,
        p1 >> p2,
        accel,
        ms
    );

    (feedrate, accel)
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    /* Steps
        1. Read beatmap and hitcircles
            a. convert curves to lines
            b. apply scaling for every point
            c. apply rotation for every point
        2. Convert to gcode paths
            a. add preliminary gcode
            b. calculate F for a given movement
            c. write movements to file
     */

    let src = OSU.parse::<OsuFile>()?;
    let binding = src.hitobjects.expect("Beatmap has no hit objects!");
    let mut hitobjects = binding.0.iter();
    let scale_fac = find_bbox_scale_factor(
        BOUNDING_BOX_SIZE,
        AREA_SIZE,
        ROTATION
    );

    println!("G90\nG0 Z0");
    let corners = [
        AREA_ORIGIN,
        (AREA_ORIGIN.0, AREA_MAX.1),
        (AREA_MAX.0, AREA_MAX.1),
        (AREA_MAX.0, AREA_ORIGIN.1)  
    ];
    for corner in corners {
        println!("G0 X{} Y{} F7500", corner.0, corner.1);
    }

    let mut total_time: f64 = 0.0;

    let mut p1 = PositionF64::new(0.0, 0.0);
    let mut p1_ms: Decimal = 0.into();
    let mut last_accel: f64 = 0.0;
    
    /*
        These are used for merging notes that share the same angle and linear
        velocity into one move. Linear velocity is plainly (dist / ms_delta).

        move_p1 is the starting point of a g-code move.
    */
    let mut move_p1: PositionF64 = p1;
    let mut move_p1_ms: Decimal = 0.into();
    let mut group_linear_vel = 0.0;
    let mut group_slope = 0.0;
    let mut group_size = 0;
    let mut dwell_state = false;

    for (count, obj) in hitobjects.enumerate() {
        let p2_ms = obj.time.get().clone().unwrap_left();
        let p1_to_p2_ms = p2_ms
            .checked_sub(p1_ms)
            .unwrap()
            .to_f64()
            .unwrap();

        let p2 = PositionF64
            ::from_position(obj.position.clone())
            .unwrap()
            .convert_to_machine_space(
                scale_fac,
                BOUNDING_BOX_SIZE,
                AREA_ORIGIN,
                AREA_SIZE,
                ROTATION
            );

        eprintln!("point {}", count);

        if count == 0 {
            println!("PAUSE");
        } else if p1 == p2 && !dwell_state {
            dwell_state = true;
        } else if p1 != p2 {
            let move_p1_to_p1_ms = p1_ms
                .checked_sub(move_p1_ms)
                .unwrap()
                .to_f64()
                .unwrap();

            let linear_vel = (p1 >> p2) / p1_to_p2_ms;
            let slope = p1 / p2;

            println!("; linear_vel {:.4}, slope {:.4}", linear_vel, slope);

            if
                linear_vel - group_linear_vel < GROUP_VEL_LENIENCY
                && slope == group_slope
            {
                println!("; !! merging point");
                group_size += 1;
            } else {
                let (feedrate, accel) = solve_feedrate_and_accel(
                    move_p1,
                    p1,
                    if group_size > 0 {
                        GROUP_ACCEL
                    } else {
                        DEFAULT_ACCEL
                    },
                    move_p1_to_p1_ms
                );

                if accel != last_accel {
                    println!("M204 S{}", accel);
                }

                println!("G0 F{:.4}\nG0 X{} Y{}", feedrate, p1.x, p1.y);

                last_accel = accel;
                group_linear_vel = linear_vel;
                group_slope = slope;
                group_size = 0;
            }

            if dwell_state {
                println!("G4 P{}", move_p1_to_p1_ms);
                dwell_state = false;
            }

            move_p1_ms = p1_ms;
            move_p1 = p1;
        }

        p1 = p2;
        p1_ms = p2_ms;
        total_time += p1_to_p2_ms;
    }

    eprintln!("total time should be {}ms", total_time);

    Ok(())
}

        /*
        println!(
            "x={} y={} Δms={} type={}",
            obj.position.x,
            obj.position.y        let move_p1_to_p1_ms = p1_ms
            .checked_sub(move_p1_ms)
            .unwrap()
            .to_f64()
            .unwrap();,
            delta_ms,
            match obj.obj_params {
                HitCircle => String::from("HitCircle"),
                Slider(s) => format!(
                    "Slider type={} points={} slides={} length={}",
                    match s.curve_type {
                        Bezier => "Bezier",
                        Centripetal => "Centripetal",
                        Linear => "Linear",
                        PerfectCircle => "PerfectCircle",
                        _ => "Other"
                    },
                    s.curve_points.len(),
                    s.slides,
                    s.length.get().clone()
                ),
                Spinner => String::from("Spinner"),
                _ => String::from("Other")
            }
        );
        */