use osu_file_parser::Position;
use rust_decimal::{Decimal, prelude::ToPrimitive};

pub type DecimalTuple = (Decimal, Decimal);

#[derive(Copy, Clone, PartialEq)]
pub struct PositionF64 {
    pub x: f64,
    pub y: f64
}

pub enum Axis {
    X,
    Y
}

impl PositionF64 {
    pub fn new(x: f64, y: f64) -> Self { Self { x, y } }

    pub fn from_tuple(a: (f64, f64)) -> Self {
        Self { x: a.0, y: a.1 }
    }

    pub fn from_position(a: Position) -> Option<Self> {
        Some(Self {
            x: a.x.get().clone().left()?.to_f64()?,
            y: a.y.get().clone().left()?.to_f64()?
        })
    }

    pub fn scale(mut self, fac: f64) -> Self {
        self.x *= fac;
        self.y *= fac;
        self
    }

    pub fn offset(mut self, x: f64, y: f64) -> Self {
        self.x += x;
        self.y += y;
        self
    }

    pub fn rotate(mut self, angle_deg: f64) -> Self {
        let theta = angle_deg * std::f64::consts::PI / 180.0;
        let sin_theta = theta.sin();
        let cos_theta = theta.cos();
        let x = self.x * cos_theta - self.y * sin_theta;
        let y = self.x * sin_theta + self.y * cos_theta;
        self.x = x;
        self.y = y;
        self
    }

    pub fn flip(mut self, axis: Axis, bisect: f64) -> Self {
        match axis {
            Axis::Y => self.x = self.x - (self.x - bisect) * 2.0,
            Axis::X => self.y = self.y - (self.y - bisect) * 2.0,
        }
        self
    }

    pub fn convert_to_machine_space(
        mut self,
        scale_fac: f64,
        bbox_size: (f64, f64),
        area_origin: (f64, f64),
        area_size: (f64, f64),
        rotation: f64
    ) -> Self {
        self.flip(Axis::X, bbox_size.1 / 2.0)
            .scale(scale_fac)
            .rotate(rotation)
            .offset(
                area_origin.0 + area_size.0,
                area_origin.1 + area_size.1 / 2.0 - bbox_size.0 * scale_fac / 2.0
            )
    }

    pub fn to_string(&self) -> String {
        format!("{:.4}, {:.4}", self.x, self.y)
    }
}

impl std::fmt::Display for PositionF64 {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.to_string());
        Ok(())
    }
}

// Add/offset
impl std::ops::Add for PositionF64 {
    type Output = Self;
    fn add(self, other: Self) -> Self::Output {
        Self {
            x: self.x + other.x,
            y: self.y + other.y
        }
    }
}

// Subtract/difference
impl std::ops::Sub for PositionF64 {
    type Output = Self;
    fn sub(self, other: Self) -> Self::Output {
        Self {
            x: self.x - other.x,
            y: self.y - other.y
        }
    }
}

// Distance
impl std::ops::Shr for PositionF64 {
    type Output = f64;
    fn shr(self, other: Self) -> Self::Output {
        let dist = other - self;
        (dist.x.powi(2) + dist.y.powi(2)).sqrt().abs()
    }
}

// Slope
impl std::ops::Div for PositionF64 {
    type Output = f64;
    fn div(self, rhs: Self) -> Self::Output {
        let diff = rhs - self;
        diff.y / diff.x
    }
}
